import numpy as np

from drymass import cfg_funcs


def test_fbool():
    assert isinstance(cfg_funcs.fbool("true"), bool)
    assert isinstance(cfg_funcs.fbool("True"), bool)
    assert isinstance(cfg_funcs.fbool("false"), bool)
    assert isinstance(cfg_funcs.fbool("False"), bool)
    assert isinstance(cfg_funcs.fbool("1"), bool)
    assert isinstance(cfg_funcs.fbool("1."), bool)
    assert isinstance(cfg_funcs.fbool("0"), bool)
    assert isinstance(cfg_funcs.fbool("0."), bool)
    assert isinstance(cfg_funcs.fbool(1.0), bool)
    assert isinstance(cfg_funcs.fbool(1), bool)
    assert isinstance(cfg_funcs.fbool(0.0), bool)
    assert isinstance(cfg_funcs.fbool(0), bool)
    assert cfg_funcs.fbool("true")
    assert cfg_funcs.fbool("True")
    assert cfg_funcs.fbool("1")
    assert cfg_funcs.fbool("1.")
    assert cfg_funcs.fbool(1.)
    assert cfg_funcs.fbool(1)
    assert not cfg_funcs.fbool("false")
    assert not cfg_funcs.fbool("False")
    assert not cfg_funcs.fbool("0.")
    assert not cfg_funcs.fbool("0")
    assert not cfg_funcs.fbool(0)
    assert not cfg_funcs.fbool(0.)

    try:
        cfg_funcs.fbool("")
    except ValueError:
        pass
    else:
        assert False, "emtpy string"


def test_fintlist():
    assert cfg_funcs.fintlist("1, 2, 3") == [1, 2, 3]
    assert cfg_funcs.fintlist("1.1, 2.1, 3.1") == [1, 2, 3]
    assert cfg_funcs.fintlist([1, 2, 3]) == [1, 2, 3]
    assert cfg_funcs.fintlist([1.1, 2.1, 3.1]) == [1, 2, 3]


def test_float_or_str():
    assert np.isnan(cfg_funcs.float_or_str("nan"))
    assert isinstance(cfg_funcs.float_or_str("1.1"), float)
    assert isinstance(cfg_funcs.float_or_str(1), float)
    assert isinstance(cfg_funcs.float_or_str(False), float)
    assert cfg_funcs.float_or_str("1.1") == 1.1
    assert cfg_funcs.float_or_str("1.1b") == "1.1b"
    assert cfg_funcs.float_or_str("1.1e-5") == 1.1e-5


def test_int_or_str():
    assert isinstance(cfg_funcs.int_or_str("1.1"), int)
    assert isinstance(cfg_funcs.int_or_str(1), int)
    assert isinstance(cfg_funcs.int_or_str(False), int)
    assert cfg_funcs.int_or_str("1.1") == 1
    assert cfg_funcs.int_or_str("1.1b") == "1.1b"
    assert cfg_funcs.int_or_str("2") == 2


def test_lcstr():
    assert cfg_funcs.lcstr("ASD") == "asd"


def test_tupletupleint():
    assert cfg_funcs.tupletupleint("") == ()
    assert cfg_funcs.tupletupleint("[(1, 2), (3, 4)]") == ((1, 2), (3, 4))
    assert cfg_funcs.tupletupleint("1, 2, 3, 4") == ((1, 2), (3, 4))
    assert cfg_funcs.tupletupleint("1.1, 2.1, 3.1, 4.1") == ((1, 2), (3, 4))
    assert cfg_funcs.tupletupleint(()) == ()
    assert cfg_funcs.tupletupleint([]) == ()
    assert cfg_funcs.tupletupleint([(1, 2), (3, 4)]) == ((1, 2), (3, 4))
    assert cfg_funcs.tupletupleint([(1.1, 2.0), (3.0, 4)]) == ((1, 2), (3, 4))


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import os
import tempfile
import shutil

import numpy as np
import qpimage

import drymass


def test_basic():
    size = 200
    x = np.arange(size).reshape(-1, 1)
    y = np.arange(size).reshape(1, -1)
    cx = 80
    cy = 120
    radius = 30
    r = np.sqrt((x - cx)**2 + (y - cy)**2)
    image = (r < radius) * 1.3
    pxsize = 1e-6
    qpi = qpimage.QPImage(data=image,
                          which_data="phase",
                          meta_data={"pixel size": pxsize})
    path = tempfile.mktemp(suffix=".h5", prefix="drymass_test_roi")
    dout = tempfile.mkdtemp(prefix="drymass_test_roi_")
    with qpimage.QPSeries(h5file=path) as qps:
        qps.add_qpimage(qpi, identifier="test")

    path_out = drymass.extract_roi(path,
                                   dir_out=dout,
                                   size_m=2 * radius * pxsize)
    with qpimage.QPSeries(h5file=path_out, h5mode="r") as qpso:
        assert len(qpso) == 1
        qpi2 = qpso.get_qpimage(0)
        assert qpi != qpi2
        assert qpi.shape != qpi2.shape

    try:
        os.remove(path)
    except OSError:
        pass
    shutil.rmtree(dout, ignore_errors=True)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import tempfile
import shutil

import numpy as np

from drymass import config_file


def test_basic():
    path = tempfile.mkdtemp(prefix="drymass_test_config_")
    cfg = config_file.ConfigFile(path=path)
    assert isinstance(cfg["bg"], dict)
    shutil.rmtree(path, ignore_errors=True)


def test_dtype():
    path = tempfile.mkdtemp(prefix="drymass_test_config_")
    cfg = config_file.ConfigFile(path=path)

    cfg["specimen"] = {"size um": 11.3}
    assert np.allclose(cfg["specimen"]["size um"], 11.3)

    cfg.set_value(section="bg",
                  key="amplitude border px",
                  value=6.3)
    assert isinstance(cfg["bg"]["amplitude border px"], int)
    assert np.allclose(cfg["bg"]["amplitude border px"], 6)
    shutil.rmtree(path, ignore_errors=True)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



from ._version import version as __version__  # noqa: F401
from .anasphere import analyze_sphere  # noqa: F401
from .converter import convert  # noqa: F401
from .extractroi import extract_roi  # noqa: F401
from . import search  # noqa: F401



import pathlib
import warnings

import numpy as np
import qpimage
from skimage.external import tifffile
import skimage.filters

from . import search
from .roi import ROIManager


# default background correction kwargs
BG_DEFAULT_KW = {"fit_offset": "mean",
                 "fit_profile": "tilt",
                 "border_perc": 5,
                 "border_px": 5,
                 }
# file names
FILE_ROI_DATA_H5 = "roi_data.h5"
FILE_ROI_DATA_TIF = "roi_data.tif"
FILE_SLICES = "roi_slices.txt"


def extract_roi(h5series, dir_out, size_m, size_var=.5, max_ecc=.7,
                dist_border=10, pad_border=40, exclude_overlap=30.,
                bg_amp_kw=BG_DEFAULT_KW, bg_amp_bin=np.nan,
                bg_pha_kw=BG_DEFAULT_KW, bg_pha_bin=np.nan,
                search_enabled=True, ret_roimgr=False):
    """Extract ROIs from a qpimage.QPSeries hdf5 file

    Parameters
    ----------
    h5series: str
        Path of qpimage.QPSeries hdf5 file
    dir_out: str
        Path to output directory
    size_m: float
        Approximate diameter of the specimen [m]
    size_var: float
        Allowed variation relative to specimen size
    max_ecc: float
        Allowed maximal eccentricity of the specimen
    dist_border: int
        Minimum distance of objects to image border [px]
    pad_border: int
        Padding of object regions [px]
    exclude_overlap: float
        Allowed distance between two objects [px]
    bg_amp_kw: dict or None
        Amplitude image background correction keyword arguments
        (see :func:`qpimage.QPImage.compute_bg`), defaults
        to `BG_DEFAULT_KW`, set to `None` to disable correction
    bg_amp_bin: float or str
        The amplitude binary threshold value or the method for binary
        threshold determination; see :mod:`skimage.filters`
        `threshold_*` methods
    bg_pha_kw: dict or None
        Phase image background correction keyword arguments
        (see :func:`qpimage.QPImage.compute_bg`), defaults
        to `BG_DEFAULT_KW`, set to `None` to disable correction
    bg_pha_bin: float or str
        The phase binary threshold value or the method for binary
        threshold determination; see :mod:`skimage.filters`
        `threshold_*` methods
    search_enabled: bool
        If True, perform automated search for ROIs using the
        parameters above. If False, extract the ROIs from `FILE_SLICES`
        and only perform background correction using the `bg_*`
        parameters.
    ret_roimgr: bool
        Return the ROIManager instance of the found ROIs
    """
    h5in = pathlib.Path(h5series)
    dout = pathlib.Path(dir_out)

    h5out = dout / FILE_ROI_DATA_H5
    imout = dout / FILE_ROI_DATA_TIF
    slout = dout / FILE_SLICES

    # Determine ROI location
    with qpimage.QPSeries(h5file=h5in, h5mode="r") as qps:
        rmgr = ROIManager(qps.identifier)
        if search_enabled:
            for ii in range(len(qps)):
                qpi = qps[ii]
                # find objects
                slices = search.search_phase_objects(
                    qpi=qpi,
                    size_m=size_m,
                    size_var=size_var,
                    max_ecc=max_ecc,
                    dist_border=dist_border,
                    pad_border=pad_border,
                    exclude_overlap=exclude_overlap)
                for jj, sl in enumerate(slices):
                    slident = "{}.{}".format(qpi["identifier"], jj)
                    rmgr.add(roislice=sl, image_index=ii,
                             roi_index=jj, identifier=slident)
            rmgr.save(slout)
        else:
            rmgr.load(slout)

    # Extract ROI images
    with qpimage.QPSeries(h5file=h5in, h5mode="r") as qps, \
            qpimage.QPSeries(h5file=h5out, h5mode="w") as qps_roi, \
            tifffile.TiffWriter(str(imout), imagej=True) as tf:
        for ii in range(len(qps)):
            # image to analyze
            qpi = qps[ii]
            # available ROIs
            rois = rmgr.get_from_image_index(ii)
            for jj, (rid, sl) in enumerate(rois):
                # Extract the ROI
                qpisl = qpi.__getitem__(sl)
                # amplitude bg correction
                if bg_amp_kw:
                    amp_mask = get_binary(
                        qpisl.amp, value_or_method=bg_amp_bin)
                    qpisl.compute_bg(which_data="amplitude",
                                     from_binary=amp_mask,
                                     **bg_amp_kw)
                # phase bg correction
                if bg_pha_kw:
                    pha_mask = get_binary(
                        qpisl.pha, value_or_method=bg_pha_bin)
                    qpisl.compute_bg(which_data="phase",
                                     from_binary=pha_mask,
                                     **bg_pha_kw)
                slident = "{}.{}".format(qpi["identifier"], jj)
                if rid != slident:
                    # This might happen if the user does not know the
                    # image identifier and builds his own `FILE_SLICES`.
                    msg = "Mismatch of slice and QPImage identifiers: " \
                          + "{} vs {}!".format(rid, slident)
                    warnings.warn(msg)
                    # override `slident` with user identifier
                    slident = rid
                qps_roi.add_qpimage(qpisl, identifier=slident)

        if len(qps_roi):
            # Write TIF
            # determine largest image
            sxmax = np.max([qq.shape[0] for qq in qps_roi])
            symax = np.max([qq.shape[1] for qq in qps_roi])
            dummy = np.zeros((2, sxmax, symax), dtype=np.float32)
            for qpir in qps_roi:
                dummy[0, :, :] = 0
                dummy[1, :, :] = 1
                res = 1 / qpir["pixel size"] * 1e-6  # use µm
                sx, sy = qpir.shape
                dummy[0, :sx, :sy] = qpir.pha
                dummy[1, :sx, :sy] = qpir.amp
                tf.save(data=dummy, resolution=(res, res, None))

    ret = h5out
    if ret_roimgr:
        ret = ret, rmgr
    return ret


def get_binary(image, value_or_method):
    if isinstance(value_or_method, str):
        method = getattr(skimage.filters, value_or_method)
        return image < method(image)
    elif np.isnan(value_or_method):
        return None
    else:
        return image < value_or_method



import argparse
import functools
import io
import os
import pathlib
import sys

import numpy as np
from PIL import Image
import qpimage
from skimage.external import tifffile

from ..anasphere import analyze_sphere
from ..converter import convert
from ..extractroi import extract_roi
from .. import config_file
from .. import definitions
from .. import plot
from .._version import version


OUTPUT_SUFFIX = "_dm"
FILE_SENSOR_WITH_ROI_IMAGE = "sensor_roi_images.tif"
FILE_SPHERE_ANALYSIS_IMAGE = "sphere_{}_{}_images.tif"


def cli_analyze_sphere(ret_data=False):
    path_out = setup_analysis()[1]
    cfg = user_complete_config_meta(path_out,
                                    meta_keys=["medium index",
                                               "pixel size um",
                                               "wavelength nm"])
    h5roi = cli_extract_roi(ret_data=True)
    print("Performing sphere analysis... ", end="", flush=True)
    # canny edge detection parameters
    edgekw = {
        "clip_rmin": cfg["sphere"]["edge clip radius min"],
        "clip_rmax": cfg["sphere"]["edge clip radius max"],
        "mult_coarse": cfg["sphere"]["edge coarse"],
        "mult_fine": cfg["sphere"]["edge fine"],
        "maxiter": cfg["sphere"]["edge iter"],
    }
    # image fitting parameters
    imagekw = {
        "crel": cfg["sphere"]["image fit range position"],
        "rrel": cfg["sphere"]["image fit range radius"],
        "nrel": cfg["sphere"]["image fit range refractive index"],
        "fix_pha_offset": cfg["sphere"]["image fix phase offset"],
        "max_iter": cfg["sphere"]["image iter"],
        "stop_dc": cfg["sphere"]["image stop delta position"],
        "stop_dr": cfg["sphere"]["image stop delta radius"],
        "stop_dn": cfg["sphere"]["image stop delta refractive index"],
        "verbose": cfg["sphere"]["image verbosity"],
    }
    h5sim = analyze_sphere(h5roiseries=h5roi,
                           dir_out=path_out,
                           r0=cfg["specimen"]["size um"] / 2 * 1e-6,
                           method=cfg["sphere"]["method"],
                           model=cfg["sphere"]["model"],
                           alpha=cfg["sphere"]["refraction increment"],
                           rad_fact=cfg["sphere"]["radial inclusion factor"],
                           edgekw=edgekw,
                           imagekw=imagekw,
                           )
    print("Done.")

    if cfg["output"]["sphere images"]:
        print("Plotting sphere images... ", end="", flush=True)
        tifout = path_out / FILE_SPHERE_ANALYSIS_IMAGE.format(
            cfg["sphere"]["method"],
            cfg["sphere"]["model"]
        )
        # plot h5series and rmgr with matplotlib
        with qpimage.QPSeries(h5file=h5roi, h5mode="r") as qps_roi, \
                qpimage.QPSeries(h5file=h5sim, h5mode="r") as qps_sim, \
                tifffile.TiffWriter(str(tifout), imagej=True) as tf:
            for ii in range(len(qps_roi)):
                qpi_real = qps_roi[ii]
                qpi_sim = qps_sim[ii]
                assert qpi_real["identifier"] in qpi_sim["identifier"]
                imio = io.BytesIO()
                plot.plot_qpi_sphere(qpi_real=qpi_real,
                                     qpi_sim=qpi_sim,
                                     path=imio,
                                     simtype=cfg["sphere"]["model"])
                imio.seek(0)
                imdat = np.array(Image.open(imio))
                tf.save(imdat)
        print("Done")


def cli_convert(ret_data=False):
    path_in, path_out = setup_analysis()
    cfg = user_complete_config_meta(path_out,
                                    meta_keys=["pixel size um",
                                               "wavelength nm"])
    print("Converting input data... ", end="", flush=True)
    meta_data = {"pixel size": cfg["meta"]["pixel size um"] * 1e-6,
                 "wavelength": cfg["meta"]["wavelength nm"] * 1e-9,
                 "medium index": cfg["meta"]["medium index"],
                 }

    holo_kw = {"filter_name": cfg["holo"]["filter name"],
               "filter_size": cfg["holo"]["filter size"],
               "sideband": cfg["holo"]["sideband"],
               }

    bg_data_amp = cfg["bg"]["amplitude data"]
    bg_data_pha = cfg["bg"]["phase data"]

    if bg_data_amp == "none":
        bg_data_amp = None

    if bg_data_pha == "none":
        bg_data_pha = None

    h5series, ds = convert(path_in=path_in,
                           dir_out=path_out,
                           meta_data=meta_data,
                           holo_kw=holo_kw,
                           bg_data_amp=bg_data_amp,
                           bg_data_pha=bg_data_pha,
                           write_tif=cfg["output"]["sensor tif data"],
                           ret_dataset=True,
                           )
    if "hologram" not in ds.storage_type:
        # remove "holo" section from configuration
        cfg.remove_section("holo")
    print("Done.")
    if ret_data:
        return h5series


def cli_extract_roi(ret_data=False):
    path_out = setup_analysis()[1]
    # cli_convert will ask for the required meta data
    h5series = cli_convert(ret_data=True)
    # get the configuration after cli_convert was run
    cfg = config_file.ConfigFile(path_out)
    print("Extracting ROIs... ", end="", flush=True)
    if cfg["bg"]["enabled"]:
        bg_amp_kw = {"fit_offset": cfg["bg"]["amplitude offset"],
                     "fit_profile": cfg["bg"]["amplitude profile"],
                     "border_perc": cfg["bg"]["amplitude border perc"],
                     "border_px": cfg["bg"]["amplitude border px"],
                     }
        bg_pha_kw = {"fit_offset": cfg["bg"]["phase offset"],
                     "fit_profile": cfg["bg"]["phase profile"],
                     "border_perc": cfg["bg"]["phase border perc"],
                     "border_px": cfg["bg"]["phase border px"],
                     }
        # compatibility for drymass.cfg files created with
        # drymass < 0.1.3 (API changed in qpimage 0.1.6).
        if bg_amp_kw["fit_profile"] == "ramp":
            bg_amp_kw["fit_profile"] = "tilt"
        if bg_pha_kw["fit_profile"] == "ramp":
            bg_pha_kw["fit_profile"] = "tilt"
    else:
        bg_amp_kw = None
        bg_pha_kw = None

    h5roi, rmgr = extract_roi(
        h5series=h5series,
        dir_out=path_out,
        size_m=cfg["specimen"]["size um"] * 1e-6,
        size_var=cfg["roi"]["size variation"],
        max_ecc=cfg["roi"]["eccentricity max"],
        dist_border=cfg["roi"]["dist border"],
        pad_border=cfg["roi"]["pad border"],
        exclude_overlap=cfg["roi"]["exclude overlap"],
        bg_amp_kw=bg_amp_kw,
        bg_amp_bin=cfg["bg"]["amplitude binary threshold"],
        bg_pha_kw=bg_pha_kw,
        bg_pha_bin=cfg["bg"]["phase binary threshold"],
        search_enabled=cfg["roi"]["enabled"],
        ret_roimgr=True,
    )
    print("Done.")
    if len(rmgr) == 0:
        print("No ROIs could be found!\n"
              + "Please try editing '{}':\n".format(cfg.path)
              + "- correct specimen size "
              + "({})\n".format(strpar(cfg, "specimen", "size um"))
              + "- increase allowed size variation "
              + "({})\n".format(strpar(cfg, "roi", "size variation"))
              + "- increase maximum allowed eccentricity "
              + "({})\n".format(strpar(cfg, "roi", "eccentricity max"))
              + "- reduce minimum distance to image border in pixels "
              + "({})\n".format(strpar(cfg, "roi", "dist border"))
              + "- reduce minimum distance inbetween ROIs "
              + "({})".format(strpar(cfg, "roi", "exclude overlap"))
              )
        sys.exit(1)

    if cfg["output"]["roi images"]:
        print("Plotting detected ROIs... ", end="", flush=True)
        tifout = path_out / FILE_SENSOR_WITH_ROI_IMAGE
        # plot h5series and rmgr with matplotlib
        with qpimage.QPSeries(h5file=h5series, h5mode="r") as qps, \
                tifffile.TiffWriter(str(tifout), imagej=True) as tf:
            for ii in range(len(qps)):
                rois = rmgr.get_from_image_index(ii)
                imio = io.BytesIO()
                plot.plot_qpi_phase(qps[ii],
                                    rois=rois,
                                    path=imio)
                imio.seek(0)
                imdat = np.array(Image.open(imio))
                tf.save(imdat)
        print("Done")

    if ret_data:
        return h5roi


@functools.lru_cache(maxsize=4)
def setup_analysis():
    print("DryMass version {}".format(version))
    parser = argparse.ArgumentParser(description='DryMass QPI analysis.')
    parser.add_argument('path', metavar='path', type=str,
                        help='Data path')
    args = parser.parse_args()
    path_in = pathlib.Path(args.path).resolve()
    if not path_in.exists():
        raise ValueError("Path '{}' does not exist!".format(path_in))
    path_out = pathlib.Path(str(path_in) + OUTPUT_SUFFIX)
    if not path_out.exists():
        os.mkdir(str(path_out))
    print("Input:  {}".format(path_in))
    print("Output: {}".format(path_out))
    return path_in, path_out


def strpar(cfg, section, key):
    return "[{}] {} = {}".format(section, key, cfg[section][key])


def user_complete_config_meta(path, meta_keys):
    cfg = config_file.ConfigFile(path)
    meta = cfg["meta"]
    for key in sorted(meta_keys):
        description = definitions.config["meta"][key][2]
        if key not in meta or np.isnan(meta[key]):
            val = input("Please enter '{}' ({}): ".format(key, description))
            cfg.set_value("meta", key, val)
    return cfg



"""Dry mass computation with radial inclusion factor

This examples illustrates the usage of the "radial inclusion factor"
which is defined in the configuration section "sphere" and used in
:py:func:`drymass.anasphere.relative_dry_mass` with the
keyword argument `rad_fact`.

The phase image is computed from two spheres whose dry masses
add up to 100pg with the larger sphere having a dry mass of 83pg.
The larger sphere is located at the center of the image which is also
used as the origin for dry mass computation. The radius of the larger
sphere is known (10µm). Thus, the corresponding radius (inner circle)
corresponds to a radial inclusion factor of 1. In DryMass, the
default radial inclusion factor is set to 1.2 (red). In some cases,
this inclusion factor must be increased or decreased depending on
whether additional information (the smaller sphere) should be included
in the dry mass computation or not.
"""
from drymass.anasphere import relative_dry_mass
import matplotlib
import matplotlib.pylab as plt
import numpy as np
import qpimage
import qpsphere

# refraction increment
alpha = .18  # [mL/g]

# general simulation parameters
medium_index = 1.333
model = "projection"
wavelength = 500e-9  # [m]
pixel_size = 1e-7  # [m]
grid_size = (400, 400)  # [px]

# sphere parameters
dry_masses = [83, 17]  # [pg]
radii = [10, 4]  # [µm]
centers = [(200, 200), (200, 340)]  # [px]

phase_data = np.zeros(grid_size, dtype=float)
for m, r, c in zip(dry_masses, radii, centers):
    # compute refractive index from dry mass
    r_m = r * 1e-6
    alpha_m3g = alpha * 1e-6
    m_g = m * 1e-12
    n = 1.333 + 3 * alpha_m3g * m_g / (4 * np.pi * (r_m**3))
    # generate example dataset
    qpi = qpsphere.simulate(radius=r_m,
                            sphere_index=n,
                            medium_index=medium_index,
                            wavelength=wavelength,
                            pixel_size=pixel_size,
                            model=model,
                            grid_size=grid_size,
                            center=c)
    phase_data += qpi.pha

qpi_sum = qpimage.QPImage(data=phase_data,
                          which_data="phase",
                          meta_data={"wavelength": wavelength,
                                     "pixel size": pixel_size,
                                     "medium index": medium_index})

# compute dry mass in dependence of radius
mass_evolution = []
mass_radii = []
for rad_fact in np.linspace(0, 2.0, 100):
    dm = relative_dry_mass(qpi=qpi_sum,
                           radius=radii[0] * 1e-6,
                           center=centers[0],
                           alpha=alpha,
                           rad_fact=rad_fact)
    mass_evolution.append(dm * 1e12)
    mass_radii.append(rad_fact)

# plot results
fig = plt.figure(figsize=(8, 3.8))
matplotlib.rcParams["image.interpolation"] = "bicubic"
# phase image
ax1 = plt.subplot(121, title="phase image [rad]")
ax1.axis("off")
map1 = ax1.imshow(qpi_sum.pha)
plt.colorbar(map1, ax=ax1, fraction=.048, pad=0.05)
# dry mass vs. inclusion factor
ax2 = plt.subplot(122, title="dry mass")
ax2.plot(mass_radii, mass_evolution)
ax2.set_ylabel("computed dry mass [pg]")
ax2.set_xlabel("radial inclusion factor")
ax2.grid()
# radius indicators
for r in [100, 180]:
    cx = centers[0][0] + .5
    cy = centers[0][1] + .5
    circle = plt.Circle((cx, cy), r,
                        color='w', fill=False, ls="dashed", lw=1, alpha=.5)
    ax1.add_artist(circle)
    ax2.axvline(r / 100, color="#404040", ls="dashed")
# add default
circle = plt.Circle((cx, cy), 120,
                    color='r', fill=False, ls="dashed", lw=1, alpha=.5)
ax1.add_artist(circle)
ax2.axvline(1.2, color="r", ls="dashed")

plt.tight_layout()
plt.show()



import pathlib
import warnings


class ROIManager(object):
    def __init__(self, identifier=None):
        self.identifier = identifier
        self.rois = []

    def __len__(self):
        return self.rois.__len__()

    def add(self, roislice, image_index, roi_index, identifier):
        # verify identifier
        if self.identifier and self.identifier not in identifier:
            msg = "`identifier` does not match dataset: {}".format(identifier)
            warnings.warn(msg, ROIManagerWarnging)
        self.rois.append((identifier, image_index, roi_index, roislice))

    def get_from_image_index(self, image_index):
        rois = [[r[0], r[3]] for r in self.rois if r[1] == image_index]
        return rois

    def load(self, path):
        path = pathlib.Path(path)
        with path.open(mode="r") as fd:
            lines = fd.readlines()
        # remove empty lines
        lines = [ll for ll in lines if ll.strip()]
        for ll in lines:
            ll = ll.strip().split("\t")
            sldata = "".join([c for c in ll[3] if c in ",0123456789"])
            sldata = sldata.replace(",,", ",").split(",")
            sldata = [int(s) for s in sldata if s]
            slice1 = slice(min(sldata[0], sldata[1]),
                           max(sldata[0], sldata[1]))
            slice2 = slice(min(sldata[2], sldata[3]),
                           max(sldata[2], sldata[3]))
            self.add(identifier=ll[0],
                     image_index=int(ll[1]),
                     roi_index=int(ll[2]),
                     roislice=(slice1, slice2)
                     )

    def save(self, path):
        path = pathlib.Path(path)
        with path.open(mode="w") as fd:
            for roi in self.rois:
                roi = [str(r) for r in roi]
                fd.write("\t".join(roi) + "\r\n")


class ROIManagerWarnging(Warning):
    pass



import numbers
import pathlib

import numpy as np
import qpformat
import qpimage
from skimage.external import tifffile


FILE_SENSOR_DATA_H5 = "sensor_data.h5"
FILE_SENSOR_DATA_TIF = "sensor_data.tif"


def convert(path_in, dir_out, meta_data={}, holo_kw={},
            bg_data_amp=None, bg_data_pha=None,
            write_tif=False, ret_dataset=False):
    """Convert experimental data to `qpimage.QPSeries` on disk

    Parameters
    ----------
    bg_data_amp, bg_data_pha: None, int, or path to file
        The background data for phase and amplitude. One of

        - `None`:
          No background data
        - `int`:
          Image index (starting at 1) of the input data set
          to use as background data
        - `str`, `pathlib.Path`:
          Path to a separate file that is used for background
          correction, relative to the directory in which `path_in`
          is located (`path_in.parent`).
    """
    path = pathlib.Path(path_in).resolve()
    dout = pathlib.Path(dir_out).resolve()

    h5out = dout / FILE_SENSOR_DATA_H5
    imout = dout / FILE_SENSOR_DATA_TIF

    ds = qpformat.load_data(path=path, meta_data=meta_data, holo_kw=holo_kw)

    if not (bg_data_amp is None and bg_data_pha is None):
        # Only set background of data set if there is
        # a background defined.
        if bg_data_amp is None:
            bgamp = np.ones(ds.get_qpimage(0).shape)
        elif isinstance(bg_data_amp, numbers.Integral):
            if bg_data_amp <= 0 or bg_data_amp > len(ds):
                msg = "Amplitude background index must be between 1 and " \
                      + "{}".format(len(ds))
                raise ValueError(msg)
            # indexing in configuration file starts at 1
            bgamp = ds.get_qpimage(bg_data_amp - 1).amp
        elif isinstance(bg_data_amp, (str, pathlib.Path)):
            bgamppath = path_in.parent / bg_data_amp
            dsbgamp = qpformat.load_data(path=bgamppath, meta_data=meta_data)
            if len(dsbgamp) != 1:
                msg = "Background correction with series data not implemented!"
                raise NotImplementedError(msg)
            else:
                bgamp = dsbgamp.get_qpimage(0).amp
        else:
            raise ValueError("Undefined bg_data_amp: {}".format(bg_data_amp))

        if bg_data_pha is None:
            bgpha = np.zeros(ds.get_qpimage(0).shape)
        elif isinstance(bg_data_amp, numbers.Integral):
            if bg_data_pha <= 0 or bg_data_pha > len(ds):
                msg = "Phase data index must be between 1 and {}".format(
                    len(ds))
                raise ValueError(msg)
            # indexing in configuration file starts at 1
            bgpha = ds.get_qpimage(bg_data_pha - 1).pha
        elif isinstance(bg_data_pha, (str, pathlib.Path)):
            bgphapath = path_in.parent / bg_data_amp
            dsbgpha = qpformat.load_data(path=bgphapath, meta_data=meta_data)
            if len(dsbgpha) != 1:
                msg = "Background correction with series data not implemented!"
                raise NotImplementedError(msg)
            else:
                bgpha = dsbgpha.get_qpimage(0).pha
        else:
            raise ValueError("Undefined bg_data_pha: {}".format(bg_data_pha))

        bg_data = qpimage.QPImage(data=(bgpha, bgamp),
                                  which_data=("phase", "amplitude"))
        ds.set_bg(bg_data)

    if h5out.exists():
        with qpimage.QPSeries(h5file=h5out, h5mode="r") as qpsr:
            if (ds.identifier == qpsr.identifier and
                    len(ds) == len(qpsr)):
                # file has same identifier and same number of qpimages
                create = False
            else:
                create = True
    else:
        create = True

    if create:
        # Write h5 data
        ds.saveh5(h5out)

    if write_tif and (create or not imout.exists()):
        # Also write tif data
        h5series2tif(h5in=h5out, tifout=imout)

    if ret_dataset:
        return h5out, ds
    else:
        return h5out


def h5series2tif(h5in, tifout):
    with qpimage.QPSeries(h5file=h5in, h5mode="r") as qps, \
            tifffile.TiffWriter(str(tifout), imagej=True) as tf:
        for ii in range(len(qps)):
            qpi = qps[ii]
            res = 1 / qpi["pixel size"] * 1e-6  # use µm
            dshape = (1, qpi.shape[0], qpi.shape[1])
            dataa = np.array(qpi.amp, dtype=np.float32).reshape(*dshape)
            datap = np.array(qpi.pha, dtype=np.float32).reshape(*dshape)
            data = np.vstack((datap, dataa))
            tf.save(data=data, resolution=(res, res, None))



import pathlib

import numpy as np
import qpimage
import qpsphere


FILE_SPHERE_DATA = "sphere_{}_{}_data.h5"
FILE_SPHERE_STAT = "sphere_{}_{}_statistics.txt"


def analyze_sphere(h5roiseries, dir_out, r0=10e-6, method="edge",
                   model="projection", edgekw={}, imagekw={},
                   alpha=.18, rad_fact=1.2):

    dir_out = pathlib.Path(dir_out).resolve()

    h5out = dir_out / FILE_SPHERE_DATA.format(method, model)
    statout = dir_out / FILE_SPHERE_STAT.format(method, model)

    with qpimage.QPSeries(h5file=h5roiseries, h5mode="r") as qps_in, \
            qpimage.QPSeries(h5file=h5out, h5mode="w") as qps_out, \
            statout.open(mode="w") as fd:

        header = ["identifier",
                  "index",
                  "radius_um",
                  "rel_dry_mass_pg",
                  "abs_dry_mass_pg",
                  "time",
                  "medium",
                  ]
        fd.write("#" + "\t".join(header) + "\r\n")

        for qpi in qps_in:
            # determine parameters
            n, r, c, qpi_sim = qpsphere.analyze(qpi,
                                                r0=r0,
                                                method=method,
                                                model=model,
                                                edgekw=edgekw,
                                                imagekw=imagekw,
                                                ret_center=True,
                                                ret_qpi=True)
            if "time" in qpi:
                qptime = qpi["time"]
            else:
                qptime = 0
            # save in txt file
            data = {
                "identifier": qpi["identifier"],
                "index": n,
                "radius_um": r * 1e6,
                "abs_dry_mass_pg": absolute_dry_mass_sphere(
                    qpi=qpi,
                    radius=r,
                    center=c,
                    alpha=alpha,
                    rad_fact=rad_fact
                ) * 1e12,
                "rel_dry_mass_pg": relative_dry_mass(
                    qpi=qpi,
                    radius=r,
                    center=c,
                    alpha=alpha,
                    rad_fact=rad_fact
                ) * 1e12,
                "time": qptime,
                "medium": qpi["medium index"]
            }
            fd.write("\t".join([str(data[k]) for k in header]) + "\r\n")
            simident = "{}:{}:{}".format(qpi["identifier"], "sim", model)
            qps_out.add_qpimage(qpi=qpi_sim, identifier=simident)

    return h5out


def absolute_dry_mass_sphere(qpi, radius, center, alpha=.18, rad_fact=1.2):
    """Compute absolute dry mass of a spherical phase object

    The absolute dry mass is computed with

    m_abs = m_rel + m_sup
    m_rel = lambda / (2*PI*alpha) * phi_tot * deltaA
    m_sup = 4*PI / (3*alpha) * radius^3 (n_med - n_PBS)

    with the vacuum wavelength `lambda`, the total phase
    retardation in the area of interest `phi_tot`, the pixel
    area `deltaA`, the refractive index of the medium `n_med`
    (stored in `qpi.meta`), and the refractive index of phosphate
    buffered saline (PBS) `n_PBS`=1.335.

    This is the *absolute* dry mass, because it takes into account
    the offset caused by the suppressed density in the phase data.

    Parameters
    ----------
    qpi: qpimage.QPImage
        QPI data
    center: tuble (x,y)
        Center of the sphere [px]
    radius: float
        Radius of the sphere [m]
    wavelength: float
        The wavelength of the light [m]
    alpha: float
        Refraction increment [mL/g]
    rad_fact: float
        Inclusion factor that scales `radius` to increase
        the area used for phase summation; if the backgound
        phase exhibits a noise phase signal, positive and
        negative contributions cancel out and `rad_fact`
        does not have an effect above a certain critical value.

    Returns
    -------
    dry_mass: float
        The absolute dry mass of the sphere [g]
    """
    dm_rel = relative_dry_mass(qpi=qpi,
                               radius=radius,
                               center=center,
                               alpha=alpha,
                               rad_fact=rad_fact)
    medium_index = qpi["medium index"]
    dm_sup = 4 / 3 * np.pi / (alpha * 1e-6) * \
        radius**3 * (medium_index - 1.335)
    return dm_rel + dm_sup


def relative_dry_mass(qpi, radius, center, alpha=.18, rad_fact=1.2):
    """Compute relative dry mass of a circular area in QPI

    The dry mass is computed with

    m_rel = lambda / (2*PI*alpha) * phi_tot * deltaA

    with the vacuum wavelength `lambda`, the total phase
    retardation in the area of interest `phi_tot`, and the pixel
    area `deltaA`.

    This is the *relative* dry mass, because it is computed relative
    to the surrounding medium (phi_tot) and not relative to water.

    Parameters
    ----------
    qpi: qpimage.QPImage
        QPI data
    center: tuble (x,y)
        Center of the area of interest [px]
    radius: float
        Radius of the area of interest [m]
    wavelength: float
        The wavelength of the light [m]
    alpha: float
        Refraction increment [mL/g]
    rad_fact: float
        Inclusion factor that scales `radius` to increase
        the area used for phase summation; if the backgound
        phase exhibits a noise phase signal, positive and
        negative contributions cancel out and `rad_fact`
        does not have an effect above a certain critical value.

    Returns
    -------
    dry_mass: float
        The relative dry mass of the object [g]
    """
    image = qpi.pha
    rincl = radius / qpi["pixel size"] * rad_fact
    wavelength = qpi["wavelength"]
    sx, sy = qpi.shape
    x = np.arange(sx).reshape(-1, 1)
    y = np.arange(sy).reshape(1, -1)
    discsq = ((x - center[0])**2 + (y - center[1])**2)
    area = discsq < rincl**2
    phi_tot = np.sum(image[area])
    # compute dry mass
    pxarea = qpi["pixel size"]**2
    # convert alpha mL/g to m³/g
    fact = 1e-6
    # [kg]
    dm = wavelength / (2 * np.pi * alpha * fact) * phi_tot * pxarea
    return dm



import numbers

import numpy as np


def fbool(value):
    """Boolean value from string or number"""
    if isinstance(value, str):
        value = value.lower()
        if value == "false":
            value = False
        elif value == "true":
            value = True
        elif value:
            value = bool(float(value))
        else:
            raise ValueError("empty string")
    else:
        value = bool(float(value))
    return value


def fintlist(alist):
    """List of integers from string or list of strings/integers"""
    outlist = []
    if isinstance(alist, str):
        # we have a string (comma-separated integers)
        alist = alist.strip().strip("[] ").split(",")
    for it in alist:
        if it:
            outlist.append(int(float(it)))
    return outlist


def float_or_str(flt_or_str):
    """Float value from string or number"""
    if isinstance(flt_or_str, str):
        flt_or_str = flt_or_str.strip()
        if flt_or_str == "nan":
            return np.nan
        else:
            try:
                value = float(flt_or_str)
            except ValueError:
                return flt_or_str
            else:
                return value
    else:
        return float(flt_or_str)


def floattuple_or_one(fti):
    """Tuple of floats or ±1 from a string or a number"""
    msg = "Expected +1, -1 or (float, float), got '{}'".format(fti)
    if isinstance(fti, str):
        try:
            fti = int(float(fti))
        except ValueError:
            raise ValueError(msg)
    if isinstance(fti, numbers.Number):
        fti = int(fti)
        if fti in [+1, -1]:
            pass
        else:
            raise ValueError(msg)
    else:
        fti = list(fti)
        if len(fti) == 2:
            try:
                fti = (float(fti[0]), float(fti[1]))
            except ValueError:
                raise ValueError(msg)
        else:
            raise ValueError(msg)
    return fti


def int_or_str(it_or_str):
    """Integer or string from a string or a number"""
    if isinstance(it_or_str, str):
        it_or_str = it_or_str.strip()
        if it_or_str.replace(".", "").isdigit():
            value = int(float(it_or_str))
        else:
            value = it_or_str
    else:
        value = int(float(it_or_str))
    return value


def lcstr(astr):
    """Convert a string to lower-case"""
    return astr.lower()


def tupletupleint(items):
    """A tuple containing x- and y- slice tuples from a string or tuple"""
    if isinstance(items, str):
        for s in " ()[]":
            items = items.replace(s, "")
        if items.strip():
            # we have a string representation of the slices
            x1, x2, y1, y2 = items.split(",")
            out = ((int(float(x1)), int(float(x2))),
                   (int(float(y1)), int(float(y2))))
        else:
            out = ()
    else:
        if items:
            (x1, x2), (y1, y2) = items
            out = ((int(float(x1)), int(float(x2))),
                   (int(float(y1)), int(float(y2))))
        else:
            out = ()
    return out


func_types = {fbool: bool,
              fintlist: list,
              lcstr: str,
              tupletupleint: tuple,
              }



"""Parse DryMass definitions.py file"""
import io
import pathlib

from docutils.statemachine import ViewList
from docutils.parsers.rst import Directive
from sphinx.util.nodes import nested_parse_with_titles
from docutils import nodes


class IncludeDirective(Directive):
    required_arguments = 1
    optional_arguments = 0

    def run(self):
        # determine path of definitions.py
        here = pathlib.Path(__file__).parent
        dfn = str(here / ".." / ".." / "drymass" / "definitions.py")

        with io.open(dfn, "r") as myfile:
            text = myfile.readlines()

        # extract desired section
        idstart = '    "{}": '.format(self.arguments[0])
        idend = '    },'
        sec_text = []
        in_sec = False
        for line in text:
            if line.startswith(idstart):
                in_sec = True
                continue
            if in_sec and line.startswith(idend):
                break
            if in_sec:
                sec_text.append(line.strip())

        rst = []

        # parse each line to build up docstring
        el_data = {}
        for line in sec_text:
            if line.startswith('"'):
                if el_data:
                    # append and parse
                    parse_element(rst, **el_data)
                    # reset data
                    el_data = {}
                el_data["name"] = line.split("#")[0].strip('": ')
            elif line.startswith("("):
                default, dtype, doc = line.split(", ", 2)
                el_data["default"] = default[1:].strip('"')
                el_data["dtype"] = dtype
                el_data["doc"] = doc.strip('"),')
            elif line.startswith("#"):
                if "notes" not in el_data:
                    el_data["notes"] = []
                el_data["notes"].append(line[1:])
        else:
            # parse last element
            parse_element(rst, **el_data)

        vl = ViewList(rst, "fakefile.rst")
        # Create a node.
        node = nodes.section()
        node.document = self.state.document
        # Parse the rst.
        nested_parse_with_titles(self.state, vl, node)
        return node.children


def parse_element(rst, name="", doc="", notes=[], default="", dtype=""):
    if default == "np.nan":
        default = "nan"

    if dtype in ["float", "int", "str"]:
        str_dtype = ":class:`{}`".format(dtype)
    else:
        str_dtype = ":func:`{0} <drymass.cfg_funcs.{0}>`".format(dtype)
    rst.append("* | **{}** = {} ({}) -- {} ".format(name, default, str_dtype, doc))
    if notes:
        rst.append("  | {}".format("".join(notes).strip()))


def setup(app):
    app.add_directive('include_definition', IncludeDirective)

    return {'version': '0.1'}   # identifies the version of our extension



from os.path import abspath, dirname
import sys

import numpy as np
import qpimage

# Add parent directory to beginning of path variable
sys.path.insert(0, dirname(dirname(abspath(__file__))))
from drymass import search  # noqa: E402


def test_basic():
    size = 200
    image = np.zeros((size, size), dtype=float)
    x = np.arange(size).reshape(-1, 1)
    y = np.arange(size).reshape(1, -1)
    cx = 80
    cy = 120
    radius = 30
    r = np.sqrt((x - cx)**2 + (y - cy)**2)
    image[r < radius] = 1.3
    rois = search.search_objects_base(image=image, size=2 * radius)
    roi = rois[0]
    assert np.allclose(roi.equivalent_diameter, 2 * radius, atol=.2, rtol=0)
    assert np.allclose(roi.centroid, (cx, cy))


def test_bg_overlap():
    size = 200
    x = np.arange(size).reshape(-1, 1)
    y = np.arange(size).reshape(1, -1)
    # 5 px between regions
    cx1 = 100
    cy1 = 100
    cx2 = 100
    cy2 = 145
    radius = 20
    r1 = np.sqrt((x - cx1)**2 + (y - cy1)**2)
    r2 = np.sqrt((x - cx2)**2 + (y - cy2)**2)
    raw_pha = (r1 < radius) * 1.3
    bg_pha = (r2 < radius) * 1.2
    # create data set
    qpi = qpimage.QPImage(data=raw_pha, bg_data=bg_pha,
                          which_data="phase",
                          meta_data={"pixel size": 1e-6})
    slices1 = search.search_phase_objects(qpi=qpi,
                                          size_m=2 * radius *
                                          qpi["pixel size"],
                                          exclude_overlap=0)
    slices2 = search.search_phase_objects(qpi=qpi,
                                          size_m=2 * radius * 1e-6,
                                          exclude_overlap=10)
    assert len(slices1) == 1
    assert len(slices2) == 0


def test_padding():
    size = 200
    x = np.arange(size).reshape(-1, 1)
    y = np.arange(size).reshape(1, -1)
    cx = 80
    cy = 120
    radius = 30
    r = np.sqrt((x - cx)**2 + (y - cy)**2)
    image = (r < radius) * 1.3
    pxsize = 1e-6
    qpi = qpimage.QPImage(data=image,
                          which_data="phase",
                          meta_data={"pixel size": pxsize})
    paddiff = 7
    [slice1] = search.search_phase_objects(qpi=qpi,
                                           size_m=2 * radius * pxsize,
                                           pad_border=0,
                                           )
    [slice2] = search.search_phase_objects(qpi=qpi,
                                           size_m=2 * radius * pxsize,
                                           pad_border=paddiff,
                                           )
    [slice3] = search.search_phase_objects(qpi=qpi,
                                           size_m=2 * radius * pxsize,
                                           pad_border=100,
                                           )
    dx1 = slice1[0].stop - slice1[0].start
    dy1 = slice1[1].stop - slice1[1].start
    dx2 = slice2[0].stop - slice2[0].start
    dy2 = slice2[1].stop - slice2[1].start
    dx3 = slice3[0].stop - slice3[0].start
    dy3 = slice3[1].stop - slice3[1].start
    assert dx2 - dx1 == 2 * paddiff
    assert dy2 - dy1 == 2 * paddiff
    assert dx3 == size
    assert dy3 == size
    assert slice3[0].start == 0
    assert slice3[1].start == 0


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import numpy as np

from .cfg_funcs import fbool, float_or_str, int_or_str, lcstr, \
    floattuple_or_one, tupletupleint


config = {
    "bg": {
        "amplitude data":
            ("none", int_or_str, "Amplitude bg correction file or index"),
            # Image indexing starts with 1.
        "amplitude offset":
            ("mean", lcstr, "Amplitude bg correction offset method"),
            # Valid values are defined in
            # :data:`qpimage.bg_estimate.VALID_FIT_OFFSETS`.
        "amplitude profile":
            ("tilt", lcstr, "Amplitude bg correction profile method"),
            # Valid values are defined in
            # :data:`qpimage.bg_estimate.VALID_FIT_PROFILES`.
        "amplitude binary threshold":
            (np.nan, float_or_str, "Binary image threshold value or method"),
            # If not *nan*, defines either a threshold for background
            # segmentation or a method in :mod:`skimage.filters`.
        "amplitude border perc":
            (10, float, "Amplitude bg border region to analyze [%]"),
        "amplitude border px":
            (5, int, "Amplitude bg border region to analyze [px]"),
        "enabled":
            (True, fbool, "Enable bg correction globally"),
            # Set to *False* when manually editing *roi_slices.txt*.
        "phase data":
            ("none", int_or_str, "Phase bg correction file or index"),
            # Image indexing starts with 1.
        "phase offset":
            ("mean", lcstr, "Phase bg correction offset method"),
            # Valid values are defined in
            # :data:`qpimage.bg_estimate.VALID_FIT_OFFSETS`.
        "phase profile":
            ("tilt", lcstr, "Phase bg correction profile method"),
            # Valid values are defined in
            # :data:`qpimage.bg_estimate.VALID_FIT_PROFILES`.
        "phase binary threshold":
            (np.nan, float_or_str, "Binary image threshold value or method"),
            # If not *nan*, defines either a threshold for background
            # segmentation or a method in :mod:`skimage.filters`.
        "phase border perc":
            (10, float, "Phase bg border region to analyze [%]"),
        "phase border px":
            (5, int, "Phase bg border region to analyze [px]"),
    },
    "holo": {
        "filter name":  # filter_name
            ("disk", str, "Filter name for sideband isolation"),
        "filter size":  # filter_size
            (1/3, float, "Filter size (fraction of the sideband frequency)"),
        "sideband":  # sideband
            (1, floattuple_or_one, "Sideband ±1 or frequency coordinates"),
    },
    "meta": {
        "medium index":
            (np.nan, float, "Refractive index of the surrounding medium"),
        "pixel size um":
            (np.nan, float, "Detector pixel size [µm]"),
        "wavelength nm":
            (np.nan, float, "Imaging wavelength [nm]"),
    },
    "output": {
        "roi images":
            (True, fbool, "Rendered phase images with ROI location"),
        "sphere images":
            (True, fbool, "Phase/Intensity images for sphere analysis"),
        "sensor tif data":
            (True, fbool, "Phase/Amplitude sensor tif data"),
    },
    "roi": {
        "dist border":
            (10, int, "Minimum distance of objects to image border [px]"),
        "eccentricity max":
            (0.7, float, "Allowed maximal eccentricity of the specimen"),
        "enabled":
            (True, fbool, "Perform automated search for ROIs"),
            # If set to *False*, the file "roi_slices.txt" must contain ROIs.
        "exclude overlap":
            (30.0, float, "Allowed distance between two objects [px]"),
        "force":
            ((), tupletupleint, "Force ROI coordinates (x1,x2,y1,y2) [px]"),
        "pad border":
            (40, int, "Padding of object regions [px]"),
        "size variation":
            (0.5, float, "Allowed variation relative to specimen size"),
    },
    "specimen": {
        "size um":
            (10, float, "Approximate diameter of the specimen [µm]"),
            # This is used as the initial value for the sphere analysis.
    },
    "sphere": {
        "edge coarse":
            (0.4, float, "Coarse edge detection filter size"),
        "edge fine":
            (0.1, float, "Fine edge detection filter size"),
        "edge clip radius min":
            (0.9, float, "Interior edge point filtering radius"),
        "edge clip radius max":
            (1.1, float, "Exterior edge point filtering radius"),
        "edge iter":
            (20, int, "Maximum number iterations for coarse edge detection"),
        "image fit range position":  # crel
            (0.05, float, "Fit interpolation range for radius"),
        "image fit range radius":  # rrel
            (0.05, float, "Fit interpolation range for radius"),
        "image fit range refractive index":  # nrel
            (0.10, float, "Fit interpolation range for refractive index"),
        "image fix phase offset":  # fix_pha_offset
            (False, fbool, "Fix the simulation background phase to zero"),
        "image iter":  # max_iter
            (100, int, "Maximum number of iterations for image fitting"),
        "image stop delta position":  # stop_dc
            (1, float, "Stopping criterion for position"),
        "image stop delta radius":  # stop_dr
            (0.0010, float, "Stopping criterion for radius"),
        "image stop delta refractive index":  # stop_dn
            (0.0005, float, "Stopping criterion for refractive index"),
        "image verbosity":  # verbose
            (1, int, "Verbosity level of image fitting algorithm"),
        "method":
            ("edge", lcstr, "Method for determining sphere parameters"),
            # Valid values are *edge* (edge-detection approach) or
            # *image* (2D phase image fitting).
        "model":
            # Valid values are defined in
            # :data:`qpsphere.models.available`. If *method=edge*, then
            # *model* must be set to *projection*. If *method=image*,
            # setting *model* to *rytov-sc* has the best trade-off between
            # accuracy and speed.
            ("projection", lcstr, "Physical sphere model"),
        "refraction increment":
            (0.18, float, "Refraction increment [mL/g]"),
        "radial inclusion factor":
            (1.2, float, "Radial inclusion factor for dry mass computation"),
    },
}



import numpy as np
from skimage.filters import threshold_li, threshold_local
from skimage.segmentation import clear_border
from skimage.morphology import label
from skimage.measure import regionprops


def approx_bg(data, filter_size=None):
    """ Approximate the image background with Gaussian convolution

    Parameters
    ----------
    data : 2d ndarray
        Data from which to compute the background
    size : float
        Approximate size of the objects on the background. The size
        of the Gaussian is heuristically determined with
        ..math::
            \sigma = 5 \cdot \texttt{size}


    Returns
    -------
    Approximate background of `data`.
    """
    if filter_size is None:
        filter_size = np.sum(data.shape) / 6

    a = np.fft.fft2(data)
    x = np.fft.fftfreq(data.shape[0]).reshape(-1, 1)
    y = np.fft.fftfreq(data.shape[1]).reshape(1, -1)

    sigma = 1 / (5 * filter_size)
    gauss = np.exp(-(x**2 + y**2) / (2 * sigma**2))
    gauss /= np.max(gauss)
    b = a * gauss
    bg = np.fft.ifft2(b)
    return bg.real


def search_objects_base(image, size=110, size_var=.5, max_ecc=.7,
                        dist_border=10, verbose=False):
    """Search objects in images

    The wrapper `search_phase_objects` implements
    a more robust (heuristic) way of finding objects.

    Parameters
    ----------
    image: 2d ndarray
        Input image
    size: float
        Approximate diameter of phase objects in pixels
    size_var: float
        Allowed variation in size (relative to `size`) for the detected
        objects
    max_ecc: float in interval [0,1)
        Maximal eccentricity of the objects. The eccentricity of a
        circle is zero. For an ellipse it is defined as
        ..math::
            e=\varepsilon=\sqrt{\frac{a^2-b^2}{a^2}}
            =\sqrt{1-\left(\frac{b}{a}\right)^2}
            =f/a
    dist_border: float
        Minimum distance of detected regions to the borders of the
        image in pixels
    verbose: bool
        If `True`, print information about ignored regions

    Returns
    -------
    rois: list of regionprops
        Found regions

    See Also
    --------
    skimage.filters.threshold_otsu: threshold for finding objects
    skimage.segmentation.clear_border: remove regions at the border
    skimage.morphology.label: identify regions in binary images
    """
    if np.allclose(image, 0, atol=1e-14, rtol=0):
        # phase images are zero
        # no regions can be found
        return []

    # apply local threshold
    block_size = ((3*size) // 2) * 2 + 1  # odd block size
    locthr = threshold_local(image, block_size=block_size)
    image = image - locthr

    # threshold image
    thresh = threshold_li(image)
    bw = image > thresh
    # label image regions
    object_labels = label(bw)

    # remove artifacts connected to image border
    clear_border(object_labels, buffer_size=int(dist_border), in_place=True)
    used_regions = []
    # Filter/draw regions
    ignored_regions = []
    for region in regionprops(object_labels):
        ds = size * size_var
        if (region.eccentricity > max_ecc or
            region.equivalent_diameter > size + ds or
                region.equivalent_diameter < size - ds):
            ignored_regions.append(region)
        else:
            used_regions.append(region)
    if verbose and len(ignored_regions) > 0:
        msg = "The following regions were ignored:\n"
        regs = []
        for reg in ignored_regions:
            regs.append(" - size: {: 7.1f}px, eccentricity: {:.1f}".format(
                reg.equivalent_diameter,
                reg.eccentricity
            )
            )
        msg += "\n".join(regs)
        print(msg)
    return used_regions


def search_phase_objects(qpi, size_m, size_var=.5, max_ecc=.7,
                         dist_border=10, pad_border=40,
                         exclude_overlap=30., verbose=False):
    """Search phase objects in quantitative phase images

    Parameters
    ----------
    qpi: qpimage.QPImage
        Quantitative phase data
    size_m: float
        Expected size of the phase objects
    size_var: float
        Allowed variation in size (relative to `size`) for the detected
        objects
    max_ecc: float in interval [0,1)
        Maximal eccentricity of the objects. The eccentricity of a
        circle is zero. For an ellipse it is defined as
        ..math::
            e=\varepsilon=\sqrt{\frac{a^2-b^2}{a^2}}
            =\sqrt{1-\left(\frac{b}{a}\right)^2}
            =f/a
    dist_border: int
        Minimum distance of detected regions to the borders of the
        image in pixels
    pad_border : int
        Pad the regions of all objects
    exclude_overlap : float
        Allowed distance in pixels between two detected regions
        (without `pad_border`)
    verbose: bool
        If `True`, print information about ignored regions

    Returns
    -------
    slices: list of slice

    Notes
    -----
    Description of the heuristic search algorithm:
    1. Search phase objects in `qpi.raw_pha` with a background
       correction computed from the gaussian-filtered image
    2. If no objects were found and `qpi.bg_pha` is nonzero,
       search objects in `qpi.pha`.
    3. Search objects in `qpi.bg_pha` and remove any overlaps
       with the previously obtained regions.

    See Also
    --------
    search_objects_base: underlying search algorithm
    approx_bg: gaussian-filtered background estimation
    """
    kwfind = {"size": size_m / qpi["pixel size"],
              "size_var": size_var,
              "max_ecc": max_ecc,
              "dist_border": dist_border,
              "verbose": verbose,
              }

    phase = qpi.raw_pha
    bgphase = qpi.bg_pha

    # Search for regions
    # First, compute regions with automatic background estimation
    bgphase_est = approx_bg(phase)
    regs = search_objects_base(phase - bgphase_est, **kwfind)
    # If this does not work, try with the provided background
    if len(regs) == 0 and not np.all(bgphase == 0):
        regs = search_objects_base(phase - bgphase, **kwfind)
    # Detect objects in the background image
    if not np.all(bgphase == 0):
        bgphasecorr = bgphase - approx_bg(bgphase)
        bgregs = search_objects_base(bgphasecorr, **kwfind)
    else:
        bgregs = []

    # Filtering
    # Filter regions that overlap with regions in the background
    delregs = []
    for rr in regs:
        for bb in bgregs:
            dst = np.sqrt((rr.centroid[0] - bb.centroid[0])**2 +
                          (rr.centroid[1] - bb.centroid[1])**2
                          )
            olap = (rr.equivalent_diameter + bb.equivalent_diameter) / 2 - dst
            if olap + exclude_overlap > 0:
                delregs.append(rr)
    for dd in delregs:
        regs.remove(dd)
    # Filter regions that overlap with regions in the image
    delregs = []
    for rr in regs:
        for bb in regs:
            if rr == bb:
                continue
            dst = np.sqrt((rr.centroid[0] - bb.centroid[0])**2 +
                          (rr.centroid[1] - bb.centroid[1])**2
                          )
            olap = (rr.equivalent_diameter + bb.equivalent_diameter) / 2 - dst
            if olap + exclude_overlap > 0:
                delregs.append(rr)
    for dd in delregs:
        if dd in regs:  # make sure `dd` is removed only once
            regs.remove(dd)
    # Create slices and pad the region sizes
    slices = []
    for re in regs:
        x1, y1, x2, y2 = re.bbox
        x1 = max(0, x1 - pad_border)
        y1 = max(0, y1 - pad_border)
        x2 = min(qpi.shape[0], x2 + pad_border)
        y2 = min(qpi.shape[1], y2 + pad_border)
        slices.append((slice(x1, x2), slice(y1, y2)))
    return slices
